﻿using System;
using System.Linq;
using System.Windows.Forms;
using CommonBehaviors.Actions;
using Styx;
using Styx.Helpers;
using Styx.Pathing;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using Styx.WoWInternals.World;
using Action = Styx.TreeSharp.Action;

namespace LazyBoxer.Helpers
{
    internal static class Movement
    {
        internal delegate WoWUnit UnitSelectionDelegate(object context);
        internal delegate bool SimpleBooleanDelegate(object context);
        internal delegate WoWPoint LocationRetriever(object context);

        private static readonly Random Random = new Random((int)DateTime.Now.Ticks);
        private static LocalPlayer Me { get { return (StyxWoW.Me); } }
        private static WoWUnit CT { get { return (StyxWoW.Me.CurrentTarget); } }
        public const double TAU = (2 * Math.PI);    // See http://tauday.com/

        /// <summary>
        ///  Creates a behavior that does nothing more than check if we're in Line of Sight of the target; and if not, move towards the target.
        /// </summary>
        /// <remarks>
        ///  Created 23/5/2011
        /// </remarks>
        /// <returns>.</returns>
        public static Composite CreateMoveToLosBehavior()
        {
            return CreateMoveToLosBehavior(ret => Me.CurrentTarget);
        }

        public static Composite CreateMoveToLosBehavior(UnitSelectionDelegate toUnit)
        {
            return new Decorator(ret => Setting.Instance.MoveToTarget.Contains("Always") && toUnit != null && toUnit(ret) != null && !toUnit(ret).IsMe && !InLineOfSpellSight(toUnit(ret)),
                new Sequence(
                    new Action(ret => Log.DebugLog("MoveToLos: moving to LoSS of {0} @ {1:F1} yds", toUnit(ret).SafeName(), toUnit(ret).Distance)),
                    new Action(ret => Navigator.MoveTo(toUnit(ret).WorldLocation)),
                    new Action(ret => StopMoving.InLosOfUnit(toUnit(ret)))
                    )
                );
        }

        private static WoWPoint Add(this WoWPoint wowPoint, double x, double y, double z)
        {
            return (new WoWPoint((wowPoint.X + x), (wowPoint.Y + y), (wowPoint.Z + z)));
        }

        private static WoWPoint AddPolarXY(this WoWPoint wowPoint, double xyHeadingInRadians, double distance, double zModifier)
        {
            return (wowPoint.Add((Math.Cos(xyHeadingInRadians) * distance), (Math.Sin(xyHeadingInRadians) * distance), zModifier));
        }

        // Finds another point near the destination.  Useful when toon is 'waiting' for something
        // (e.g., boat, mob repops, etc). This allows multiple people running
        // the same profile to not stand on top of each other while waiting for
        // something.
        // Borrowed from QuestBehaviors default with HB
        public static WoWPoint LocationNearHere(this WoWPoint location, double maxRadius)
        {
            const int CYLINDER_LINE_COUNT = 12;
            const int MAX_TRIES = 50;
            const double SAFE_DISTANCE_BUFFER = 1.75;

            WoWPoint candidateDestination = location;
            int tryCount;

            // Most of the time we'll find a viable spot in less than 2 tries...
            // However, if you're standing on a pier, or small platform a
            // viable alternative may take 10-15 tries--its all up to the
            // random number generator.
            for (tryCount = MAX_TRIES; tryCount > 0; --tryCount)
            {
                bool[] hitResults;
                WoWPoint[] hitPoints;
                WorldLine[] traceLines = new WorldLine[CYLINDER_LINE_COUNT + 1];

                candidateDestination = location.AddPolarXY((TAU * Random.NextDouble()), (maxRadius * Random.NextDouble()), 0.0);

                // Build set of tracelines that can evaluate the candidate destination --
                // We build a cone of lines with the cone's base at the destination's 'feet',
                // and the cone's point at maxRadius over the destination's 'head'.  We also
                // include the cone 'normal' as the first entry.

                // 'Normal' vector
                int index = 0;
                traceLines[index].Start = candidateDestination.Add(0.0, 0.0, maxRadius);
                traceLines[index].End = candidateDestination.Add(0.0, 0.0, -maxRadius);

                // Cylinder vectors
                for (double turnFraction = 0.0; turnFraction < TAU; turnFraction += (TAU / CYLINDER_LINE_COUNT))
                {
                    ++index;
                    WoWPoint circlePoint = candidateDestination.AddPolarXY(turnFraction, SAFE_DISTANCE_BUFFER, 0.0);
                    traceLines[index].Start = circlePoint.Add(0.0, 0.0, maxRadius);
                    traceLines[index].End = circlePoint.Add(0.0, 0.0, -maxRadius);
                }


                // Evaluate the cylinder...
                // The result for the 'normal' vector (first one) will be the location where the
                // destination meets the ground.  Before this MassTrace, only the candidateDestination's
                // X/Y values were valid.
                GameWorld.MassTraceLine(traceLines.ToArray(),
                                        GameWorld.CGWorldFrameHitFlags.HitTestGroundAndStructures,
                                        out hitResults,
                                        out hitPoints);

                candidateDestination = hitPoints[0];    // From 'normal', Destination with valid Z coordinate


                // Sanity check...
                // We don't want to be standing right on the edge of a drop-off (say we'e on
                // a plaform or pier).  If there is not solid ground all around us, we reject
                // the candidate.  Our test for validity is that the walking distance must
                // not be more than 20% greater than the straight-line distance to the point.
                int viableVectorCount = hitPoints.Sum(point => ((Me.Location.SurfacePathDistance(point) < (Me.Location.Distance(point) * 1.20))
                                                                      ? 1
                                                                      : 0));

                if (viableVectorCount < (CYLINDER_LINE_COUNT + 1))
                { continue; }

                // If new destination is 'too close' to our current position, try again...
                if (Me.Location.Distance(candidateDestination) <= SAFE_DISTANCE_BUFFER)
                { continue; }

                break;
            }

            // If we exhausted our tries, just go with simple destination --
            if (tryCount <= 0)
            { candidateDestination = location; }

            return (candidateDestination);
        }


        private static double SurfacePathDistance(this WoWPoint start, WoWPoint destination)
        {
            WoWPoint[] groundPath = Navigator.GeneratePath(start, destination) ?? new WoWPoint[0];

            // We define an invalid path to be of 'infinite' length
            if (groundPath.Length <= 0)
            { return (double.MaxValue); }


            double pathDistance = start.Distance(groundPath[0]);

            for (int i = 0; i < (groundPath.Length - 1); ++i)
            { pathDistance += groundPath[i].Distance(groundPath[i + 1]); }

            return (pathDistance);
        }

        /// <summary>
        /// true if Target in line of spell sight AND a spell hasn't just failed due
        /// to a line of sight error.  This test is required for the unit we are moving
        /// towards because WoWUnit.InLineOfSpellSight will return true while the
        /// WOW Game Client fails the spell cast.  See EventHandler.cs for setting
        /// LastLineOfSightError
        /// 
        /// Only use this for unit we are moving towards.  Not to be used when checking
        /// ability to cast spells on various mobs
        /// </summary>
        /// <param name="unit">target we are moving towards</param>
        /// <returns></returns>
        public static bool InLineOfSpellSight(WoWUnit unit)
        {
            if (unit.InLineOfSpellSight)
            {
                /*
                if ((DateTime.Now - EventHandlers.LastLineOfSightFailure).TotalMilliseconds < 1000)
                {
                    if (unit.IsWithinMeleeRange)
                    {
                        Logger.WriteDebug(Color.White, "InLineOfSpellSight: last LoS error < 1 sec but in melee range, pretending we are in LoS");
                        return true;
                    }

                    Logger.WriteDebug(Color.White, "InLineOfSpellSight: last LoS error < 1 sec, pretending still not in LoS");
                    return false;
                }
                 */

                return true;
            }

            return false;
        }

        /// <summary>
        ///   Creates the ensure movement stopped behavior. if no range specified, will stop immediately.  if range given, will stop if within range yds of current target
        /// </summary>
        /// <remarks>
        ///   Created 5/1/2011.
        /// </remarks>
        /// <returns>.</returns>
        public static Composite CreateEnsureMovementStoppedBehavior(float range = float.MaxValue, UnitSelectionDelegate onUnit = null, string reason = null)
        {
                if (range == float.MaxValue)
                {
                    return new Decorator( ret => Setting.Instance.MoveToTarget.Contains("Always") && StyxWoW.Me.IsMoving,
                        new Sequence(
                            new Action(ret => Log.DebugLog("EnsureMovementStopped: stopping! {0}", reason ?? "")),
                            new Action(ret => StopMoving.Now())
                            )
                        );
                }

                if (onUnit == null)
                    onUnit = del => StyxWoW.Me.CurrentTarget;

                return new Decorator( ret => Setting.Instance.MoveToTarget.Contains("Always")
                           && StyxWoW.Me.IsMoving && (onUnit(ret) == null || onUnit(ret).Distance < range),
                    new Sequence(
                        new Action( ret => Log.DebugLog("EnsureMovementStopped: stopping because {0}", onUnit(ret) == null ? "No CurrentTarget" : string.Format("target @ {0:F1} yds, stop range: {1:F1}",onUnit(ret).Distance, range))),
                        new Action(ret => StopMoving.Now())
                        )
                    );
        }

        /// <summary>
        /// Creates ensure movement stopped if within melee range behavior.
        /// </summary>
        /// <returns></returns>
        public static Composite CreateEnsureMovementStoppedWithinMelee()
        {
            return new Decorator(ret => Setting.Instance.MoveToTarget.Contains("Always") && StyxWoW.Me.IsMoving && InMoveToMeleeStopRange(StyxWoW.Me.CurrentTarget),
                new Sequence(
                    new Action(ret => Log.DebugLog("EnsureMovementStoppedWithinMelee: stopping because {0}", !StyxWoW.Me.GotTarget ? "No CurrentTarget" : string.Format("target at {0:F1} yds", StyxWoW.Me.CurrentTarget.Distance))),
                    new Action(ret => StopMoving.Now())
                    )
                );
        }

        /// <summary>
        ///   Creates a behavior that does nothing more than check if we're facing the target; and if not, faces the target. (Uses a hard-coded 70degree frontal cone)
        /// </summary>
        /// <remarks>
        ///   Created 5/1/2011.
        /// </remarks>
        /// <returns>.</returns>
        public static Composite CreateFaceTargetBehavior(float viewDegrees = 70f, bool waitForFacing = true)
        {
            return CreateFaceTargetBehavior(ret => StyxWoW.Me.CurrentTarget, viewDegrees, waitForFacing);
        }

        public static Composite CreateFaceTargetBehavior(UnitSelectionDelegate toUnit, float viewDegrees = 70f, bool waitForFacing = true)
        {
            return new Decorator(
                ret => Setting.Instance.MoveToTarget.Contains("Always")
                    && toUnit != null && toUnit(ret) != null
                    && !StyxWoW.Me.IsMoving
                    && !toUnit(ret).IsMe
                    && !StyxWoW.Me.IsSafelyFacing(toUnit(ret), viewDegrees),
                new Action(ret =>
                {
                    toUnit(ret).Face();

                    // check only if standard facing for wait, not value given by viewDegrees
                    // .. this optimizes for combat... cases where facing within viewDegrees
                    // .. must be enforced should use a different behavior

                    if (waitForFacing && !StyxWoW.Me.IsSafelyFacing(toUnit(ret)))
                        return RunStatus.Success;

                    return RunStatus.Failure;
                })
                );
        }

        /// <summary>
        ///   Creates a move to target behavior. Will return RunStatus.Success if it has reached the location, or stopped in range. Best used at the end of a rotation.
        /// </summary>
        /// <remarks>
        ///   Created 5/1/2011.
        /// </remarks>
        /// <param name = "stopInRange">true to stop in range.</param>
        /// <param name = "range">The range.</param>
        /// <returns>.</returns>
        public static Composite CreateMoveToTargetBehavior(bool stopInRange, float range)
        {
            return CreateMoveToTargetBehavior(stopInRange, range, ret => StyxWoW.Me.CurrentTarget);
        }

        /// <summary>
        ///   Creates a move to target behavior. Will return RunStatus.Success if it has reached the location, or stopped in range. Best used at the end of a rotation.
        /// </summary>
        /// <remarks>
        ///   Created 5/1/2011.
        /// </remarks>
        /// <param name = "stopInRange">true to stop in range.</param>
        /// <param name = "range">The range.</param>
        /// <param name="onUnit">The unit to move to.</param>
        /// <returns>.</returns>
        public static Composite CreateMoveToTargetBehavior(bool stopInRange, float range, UnitSelectionDelegate onUnit)
        {
            return
                new Decorator( ret => onUnit != null && onUnit(ret) != null && onUnit(ret) != StyxWoW.Me && !StyxWoW.Me.IsCasting,
                    CreateMoveToLocationBehavior(ret => onUnit(ret).WorldLocation, stopInRange, ret => onUnit(ret).SpellRange(range)));
        }

        public static Composite CreateMoveToUnitBehavior(float range, UnitSelectionDelegate onUnit)
        {
            return CreateMoveToUnitBehavior(onUnit, range);
        }

        public static Composite CreateMoveToUnitBehavior(UnitSelectionDelegate onUnit, float range, float stopAt = float.MinValue)
        {
            return new Decorator(
                ret => Setting.Instance.MoveToTarget.Contains("Always") && onUnit != null && onUnit(ret) != null && onUnit(ret).Distance > range,
                new Sequence(
                    new Action(ret => Log.DebugLog("MoveToUnit: moving within {0:F1} yds of {1} @ {2:F1} yds", range, onUnit(ret).SafeName(), onUnit(ret).Distance)),
                    new Action(ret => Navigator.MoveTo(onUnit(ret).WorldLocation)),
                    new Action(ret => StopMoving.InRangeOfUnit(onUnit(ret), stopAt == float.MinValue ? range : stopAt)),
                    new ActionAlwaysFail()
                    )
                );
        }

#if USE_OLD_VERSION
        /// <summary>
        ///   Creates a move to melee range behavior. Will return RunStatus.Success if it has reached the location, or stopped in range. Best used at the end of a rotation.
        /// </summary>
        /// <remarks>
        ///   Created 5/1/2011.
        /// </remarks>
        /// <param name = "stopInRange">true to stop in range.</param>
        /// <param name = "range">The range.</param>
        /// <returns>.</returns>
        public static Composite CreateMoveToMeleeBehavior(bool stopInRange)
        {
            return CreateMoveToMeleeBehavior(ret => StyxWoW.Me.CurrentTarget.WorldLocation, stopInRange);
        }
#else

        /// <summary>
        /// Creates a move to melee range behavior.  Tests .IsWithinMeleeRange so we know whether WoW thinks
        /// we are within range, which is more important than our distance calc.  For players keep moving 
        /// until 2 yds away so we stick to them in pvp
        /// </summary>
        /// <param name="stopInRange"></param>
        /// <returns></returns>
        public static Composite CreateMoveToMeleeBehavior(bool stopInRange)
        {
#if OLD_MELEE_MOVE
            return new Decorator(
                ret => Setting.Instance.MoveToTarget.Contains("Always"),
                new PrioritySelector(
                    new Decorator(
                        ret => stopInRange && InMoveToMeleeStopRange(StyxWoW.Me.CurrentTarget),
                        new PrioritySelector(
                            CreateEnsureMovementStoppedWithinMelee(),
                            new Action(ret => RunStatus.Success)
                            )
                        ),
                    new Decorator(
                        ret => StyxWoW.Me.CurrentTarget != null && StyxWoW.Me.CurrentTarget.IsValid,
                        new Sequence(
                            new Action(ret => Log.DebugLog("MoveToMelee: towards {0} @ {1:F1} yds", StyxWoW.Me.CurrentTarget.SafeName(), StyxWoW.Me.CurrentTarget.Distance)),
                            new Action(ret => Navigator.MoveTo(StyxWoW.Me.CurrentTarget.WorldLocation)),
                            new Action(ret => StopMoving.InMeleeRangeOfUnit(StyxWoW.Me.CurrentTarget, and => InMoveToMeleeStopRange(StyxWoW.Me.CurrentTarget))),
                            new ActionAlwaysFail()
                            )
                        )
                    )
                );
#else
            return new Decorator(
                ret => Setting.Instance.MoveToTarget.Contains("Always") && StyxWoW.Me.CurrentTarget != null && !StyxWoW.Me.CurrentTarget.IsWithinMeleeRange,
                new Sequence(
                    new Action(ret =>
                    {
                        MoveResult res = Navigator.MoveTo(StyxWoW.Me.CurrentTarget.WorldLocation);
                        Log.DebugLog("MoveToMelee({0}): towards {1} @ {2:F1} yds", res.ToString(), StyxWoW.Me.CurrentTarget.SafeName(), StyxWoW.Me.CurrentTarget.Distance);
                    }),
                    new Action(ret => StopMoving.InMeleeRangeOfUnit(StyxWoW.Me.CurrentTarget)),
                    new ActionAlwaysFail()
                    )
                );
#endif
        }

        public static bool InMoveToMeleeStopRange(WoWUnit unit)
        {
            if (unit == null || !unit.IsValid)
                return true;

            if (unit.IsPlayer)
            {
                if (unit.DistanceSqr < (2 * 2))
                    return true;

            }
            else
            {
                float preferredDistance = unit.MeleeDistance() - 1f;
                if (unit.Distance <= preferredDistance)
                    return true;
            }

            return !unit.IsMoving && unit.IsWithinMeleeRange;
        }
#endif


        #region Move Behind

        /// <summary>
        ///   Creates a move behind target behavior. If it cannot fully navigate will move to target location
        /// </summary>
        /// <remarks>
        ///   Created 2/12/2011.
        /// </remarks>
        /// <returns>.</returns>
        public static Composite CreateMoveBehindTargetBehavior()
        {
            return CreateMoveBehindTargetBehavior(ret => true);
        }

        /// <summary>
        ///   Creates a move behind target behavior. If it cannot fully navigate will move to target location
        /// </summary>
        /// <remarks>
        ///   Created 2/12/2011.
        /// </remarks>
        /// <param name="requirements">Aditional requirments.</param>
        /// <returns>.</returns>
        public static Composite CreateMoveBehindTargetBehavior(SimpleBooleanDelegate requirements)
        {
            return new Decorator(ret =>
                {
                    if (Setting.Instance.MoveToTarget.Contains("Never") || !requirements(ret) || StyxWoW.Me.IsCasting || Groups.MeIsTank) return false;
                    var currentTarget = StyxWoW.Me.CurrentTarget;
                    if (currentTarget == null || currentTarget.MeIsSafelyBehind || !currentTarget.IsAlive)// || BossList.AvoidRearBosses.Contains(currentTarget.Entry))
                        return false;
                    return currentTarget.Stunned || currentTarget.CurrentTarget != StyxWoW.Me;
                },
                new PrioritySelector(ctx => CalculatePointBehindTarget(),
                    new Decorator(req => Navigator.CanNavigateFully(StyxWoW.Me.WorldLocation, (WoWPoint)req, 4),
                        new Sequence(
                            new Action(ret => Log.DebugLog("MoveBehind: behind {0} @ {1:F1} yds", StyxWoW.Me.CurrentTarget.SafeName(), StyxWoW.Me.CurrentTarget.Distance)),
                            new Action(behindPoint => Navigator.MoveTo((WoWPoint)behindPoint)),
                            new Action(behindPoint => StopMoving.AtLocation((WoWPoint)behindPoint))
                            )
                        )
                    )
                );
        }

        public static WoWPoint CalculatePointBehindTarget()
        {
            float facing = StyxWoW.Me.CurrentTarget.Rotation;
            facing += WoWMathHelper.DegreesToRadians(180); // was 150 ?
            facing = WoWMathHelper.NormalizeRadian(facing);

            return StyxWoW.Me.CurrentTarget.WorldLocation.RayCast(facing, StyxWoW.Me.CurrentTarget.MeleeRange() - 2f);
        }

        #endregion

        #region Root Move To Location

        /// <summary>
        ///   Creates a move to location behavior. Will return RunStatus.Success if it has reached the location, or stopped in range. Best used at the end of a rotation.
        /// </summary>
        /// <remarks>
        ///   Created 5/1/2011.
        /// </remarks>
        /// <param name = "location">The location.</param>
        /// <param name = "stopInRange">true to stop in range.</param>
        /// <param name = "range">The range.</param>
        /// <returns>.</returns>
        public static Composite CreateMoveToLocationBehavior(LocationRetriever location, bool stopInRange, DynamicRangeRetriever range)
        {
            // Do not fuck with this. It will ensure we stop in range if we're supposed to.
            // Otherwise it'll stick to the targets ass like flies on dog shit.
            // Specifying a range of, 2 or so, will ensure we're constantly running to the target. Specifying 0 will cause us to spin in circles around the target
            // or chase it down like mad. (PVP oriented behavior)
            return
                new Decorator(ret => Setting.Instance.MoveToTarget.Contains("Always"),
                    
                    new PrioritySelector(
                        new Decorator(ret => stopInRange && StyxWoW.Me.WorldLocation.Distance(location(ret)) < range(ret),
                // Give it a little more than 1/2 a yard buffer to get it right. CTM is never 'exact' on where we land. So don't expect it to be.
                            new PrioritySelector(
                                CreateEnsureMovementStoppedBehavior(),
                // In short; if we're not moving, just 'succeed' here, so we break the tree.
                                new Action(ret => RunStatus.Success)
                                )
                            ),
                        new Action(ret => Navigator.MoveTo(location(ret))),
                        new Action(ret => StopMoving.InRangeOfLocation(location(ret), range(ret)))
                        ));
        }

        #endregion

    }

    public static class StopMoving
    {
        internal static StopType Type { get; set; }
        internal static WoWPoint Point { get; set; }
        internal static WoWUnit Unit { get; set; }
        internal static double Range { get; set; }
        internal static SimpleBooleanDelegate StopNow { get; set; }
        internal delegate WoWUnit UnitSelectionDelegate(object context);

        public delegate bool SimpleBooleanDelegate(object context);
        internal delegate WoWPoint LocationRetriever(object context);

        public enum StopType
        {
            None = 0,
            AsSoonAsPossible,
            Location,
            RangeOfLocation,
            RangeOfUnit,
            MeleeRangeOfUnit,
            LosOfUnit
        }

        static StopMoving()
        {
            Clear();
        }

        public static void Clear()
        {
            Set(StopType.None, null, WoWPoint.Empty, 0, stop => false, null);
        }

        public static void Pulse()
        {
            if (Type == StopType.None)
                return;

            bool stopMovingNow;
            try
            {
                stopMovingNow = StopNow(null);
            }
            catch
            {
                stopMovingNow = true;
            }

            if (stopMovingNow)
            {
                if (!StyxWoW.Me.IsMoving)
                    Log.DebugLog("StopMoving: character already stopped, clearing stop {0} request", Type);
                else
                {
                    Navigator.PlayerMover.MoveStop();

                    string line = string.Format("StopMoving: {0}", Type);
                    if (Type == StopType.Location)
                        line += string.Format(", within {0:F1} yds of {1}", StyxWoW.Me.WorldLocation.Distance(Point), Point);
                    else if (Type == StopType.RangeOfLocation)
                        line += string.Format(", within {0:F1} yds of {1} @ {2:F1} yds", Range, Point, StyxWoW.Me.WorldLocation.Distance(Point));
                    else if (Unit == null || !Unit.IsValid)
                        line += ", unit == null";
                    else if (Type == StopType.LosOfUnit)
                        line += string.Format(", have LoSS of {0} @ {1:F1} yds", Unit.SafeName(), Unit.Distance);
                    else if (Type == StopType.MeleeRangeOfUnit)
                        line += string.Format(", within melee range of {0} @ {1:F1} yds", Unit.SafeName(), Unit.Distance);
                    else if (Type == StopType.RangeOfUnit)
                        line += string.Format(", within {0:F1} yds of {1} @ {2:F1} yds", Range, Unit.SafeName(), Unit.Distance);

                    Log.DebugLog(line);
                }

                Clear();
            }
        }

        private static void Set(StopType type, WoWUnit unit, WoWPoint pt, double range, SimpleBooleanDelegate stop, SimpleBooleanDelegate and)
        {
            if (Setting.Instance.MoveToTarget.Contains("Never"))
                return;

            Type = type;
            Unit = unit;
            Point = pt;
            Range = range;

            if (and == null)
                and = ret => true;

            StopNow = ret => stop(ret) && and(ret);
        }

        public static void AtLocation(WoWPoint pt, SimpleBooleanDelegate and = null)
        {
            Set(StopType.Location, null, pt, 0, at => StyxWoW.Me.WorldLocation.Distance(pt) <= 1, and);
        }

        public static void InRangeOfLocation(WoWPoint pt, double range, SimpleBooleanDelegate and = null)
        {
            Set(StopType.RangeOfLocation, null, pt, range, at => StyxWoW.Me.WorldLocation.Distance(pt) <= range, and);
        }

        public static void InRangeOfUnit(WoWUnit unit, double range, SimpleBooleanDelegate and = null)
        {
            Set(StopType.RangeOfUnit, unit, WoWPoint.Empty, range, at => Unit == null || !Unit.IsValid || Unit.Distance <= range, and);
        }

        public static void InMeleeRangeOfUnit(WoWUnit unit, SimpleBooleanDelegate and = null)
        {
            Set(StopType.RangeOfUnit, unit, WoWPoint.Empty, 0, at => Unit == null || !Unit.IsValid || Movement.InMoveToMeleeStopRange(Unit), and);
        }

        public static void InLosOfUnit(WoWUnit unit, SimpleBooleanDelegate and = null)
        {
            Set(StopType.LosOfUnit, unit, WoWPoint.Empty, 0, at => Unit == null || !Unit.IsValid || Movement.InLineOfSpellSight(Unit), and);
        }

        public static void Now()
        {
            Clear();
            if (StyxWoW.Me.IsMoving)
            {
                Log.DebugLog("StopMoving.Now: character already stopped, clearing stop {0} request", Type);
                Navigator.PlayerMover.MoveStop();
            }
        }

        public static void AsSoonAsPossible(SimpleBooleanDelegate and = null)
        {
            Set(StopType.AsSoonAsPossible, null, WoWPoint.Empty, 0, at => true, and);
        }
    }

    public delegate WoWPoint LocationRetriever(object context);

    public delegate float DynamicRangeRetriever(object context);



    #region Local Classes for No Movement Providers - NAVIGATION

    class NoNavigation : INavigationProvider
    {
        public bool CanNavigateFully(WoWPoint from, WoWPoint to, int maxHops) { return true; }
        public bool Clear() { return true; }
        public WoWPoint[] GeneratePath(WoWPoint from, WoWPoint to) { return new WoWPoint[] { new WoWPoint(from.X, from.Y, from.Z) }; }
        public MoveResult MoveTo(WoWPoint location) { return MoveResult.Moved; }
        public float PathPrecision { get; set; }
        public IStuckHandler StuckHandler { get; set; }

        public NoNavigation()
        {
            StuckHandler = new NoStuckHandling();
        }
    }

    class NoPlayerMovement : IPlayerMover
    {
        public void Move(WoWMovement.MovementDirection direction)   { }
        public void MoveStop() { }
        public void MoveTowards(WoWPoint location) { }
    }

    class NoStuckHandling : IStuckHandler
    {
        public bool IsStuck() { return false; }
        public void Reset() { }
        public void Unstick() { }
    }




    // Click To Move player navigation
    class CTMNavigation : INavigationProvider
    {
        public bool CanNavigateFully(WoWPoint from, WoWPoint to, int maxHops)
        {
            //Log.Info("==== CTM Movement: can navigate");
            return true;
        }

        public bool Clear() { return true; }
        public WoWPoint[] GeneratePath(WoWPoint from, WoWPoint to) { return new WoWPoint[] { new WoWPoint(from.X, from.Y, from.Z) }; }
        public MoveResult MoveTo(WoWPoint location)
        {
            //Log.Info("==== CTM Movement: move to");
            WoWMovement.ClickToMove(location);
            return MoveResult.Moved;
        }
        public float PathPrecision { get; set; }
        public IStuckHandler StuckHandler { get; set; }

        public CTMNavigation()
        {
            StuckHandler = new NoStuckHandling();
        }
    }

    class CTMPlayerMovement : IPlayerMover
    {
        public void Move(WoWMovement.MovementDirection direction)
        {
            switch (direction)
            {
                    case WoWMovement.MovementDirection.Backwards:
                    KeyboardManager.PressKey((char)Keys.Down);
                    break;

                    case WoWMovement.MovementDirection.Forward:
                    KeyboardManager.PressKey((char)Keys.Up);
                    break;

                    case WoWMovement.MovementDirection.JumpAscend:
                    KeyboardManager.PressKey((char)Keys.Space);
                    break;

                    case WoWMovement.MovementDirection.Descend:
                    KeyboardManager.PressKey((char)Keys.S);
                    break;

                    case WoWMovement.MovementDirection.StrafeLeft:
                    KeyboardManager.PressKey((char)Keys.Q);
                    break;

                    case WoWMovement.MovementDirection.StrafeRight:
                    KeyboardManager.PressKey((char)Keys.E);
                    break;

                    case WoWMovement.MovementDirection.TurnLeft:
                    KeyboardManager.PressKey((char)Keys.Left);
                    break;

                    case WoWMovement.MovementDirection.TurnRight:
                    KeyboardManager.PressKey((char)Keys.Right);
                    break;

                    case WoWMovement.MovementDirection.None:
                    KeyboardManager.ReleaseKey((char)Keys.Down);
                    KeyboardManager.ReleaseKey((char)Keys.Left);
                    KeyboardManager.ReleaseKey((char)Keys.Right);
                    KeyboardManager.ReleaseKey((char)Keys.Up);
                    KeyboardManager.ReleaseKey((char)Keys.Space);
                    KeyboardManager.ReleaseKey((char)Keys.S);
                    KeyboardManager.ReleaseKey((char)Keys.Q);
                    KeyboardManager.ReleaseKey((char)Keys.D);
                    KeyboardManager.ReleaseKey((char)Keys.A);
                    KeyboardManager.ReleaseKey((char)Keys.W);
                    KeyboardManager.ReleaseKey((char)Keys.E);
                    break;

            }
        }

        public void MoveStop()
        {
            //Log.Info("==== CTM Movement: move stop");
            KeyboardManager.KeyUpDown((char) Keys.Down);
        }

        public void MoveTowards(WoWPoint location)
        {
            //Log.Info("==== CTM Movement: move towards");
            WoWMovement.ClickToMove(location);
        }
    }


        #endregion

}
